unit fr_3D_Glyph;

{$mode objfpc}{$H+}

interface

uses
  Graphics,Classes, SysUtils,EasyLazFreeType, TTObjs, TTTypes,
  bezier,
  fr_typen,
  BeRoTriangulation;

type

TPolygon_punkt=class;
TGlyph_polygon=class;
T3D_Glyph=class;

TKB_eigenschaft=(Konturpunkt,Bezierpunkt);
TSEZ_eigenschaft=(Startpunkt,Endpunkt,Zwischenpunkt);

TDrehrichtung=(rechts,links);

// Bounding-Box-typ
Tbox=record
 min:punkt3d;
 max:punkt3d;
end;

// Dreieck-Objekt
TDreieck=class
 A,B,C,nv:Punkt3d;
 procedure berechne_normalvektor;
end;

// Polygon-Punkte mit Eigenschaften
TPolygon_punkt=class
 x,y:frfloat;
 KB_eigenschaft:TKB_eigenschaft; // Koturpunkt oder Bezierpunkt
 SEZ_eigenschaft:TSEZ_eigenschaft; // Startpunkt,Endpunkt,Zwischenpunkt
end;

// Polygone die in einem Glyph enthalten sind , können auch mehrere sein
TGlyph_polygon=class
 punkte_liste:tlist; // TPolygon_punkt
 Drehrichtung:TDrehrichtung;
 constructor create;
 destructor destroy; override;
 procedure zeichne(canvas:tcanvas;scale:frfloat;offset_x,offset_y:integer; // Auf Canvas zeichnen
                   punktgroesse:integer;punkte_zeichnen:boolean);
 procedure rotiere_punkteliste_rechts; // Rotation der Polygone , wenn der Startpunkt eine Bezierpunkt ist
 function glaette(aufloesung:integer):TGlyph_polygon; // Bezierpunkte auflösen
 procedure berechne_drehrichtung;
end;

// 3D-Objekt eines Glyph's
T3D_Glyph=class
 _glyph:PGlyph;
 bbox:trect;
 box:Tbox;
 Dreiecklist2D:tlist; //TDreieck platt in 2D , Z-Koordinate = 0 , nur den vorderen Deckel
 Dreiecklist3D:tlist; //TDreieck Körper extrudiert
 anzahl_polygone:integer;
 Grob_Polys:tlist; //TGlyph_polygon; // Orginal Polygone wie in Fontdatei
 Fein_Polys:tlist; //TGlyph_polygon; // Konturpolygon nach Auflösung der Bezierstützpunkte
 constructor create(const glyph:TFreeTypeGlyph);
 destructor destroy; override;
 procedure berechne_box;
 function mach_bero_polys:TBeRoTriangulationPolygons;
 procedure zeichne_grob_polys(canvas:tcanvas;
                              scalefaktor:frfloat;
                              offset_x,offset_y:integer;
                              punkte_ja_nein:boolean);
 procedure zeichne_fein_polys(canvas:tcanvas;
                              scalefaktor:frfloat;
                              offset_x,offset_y:integer;
                              punkte_ja_nein:boolean);
 procedure zeichne_dreiecke_2D(canvas:tcanvas;
                               scalefaktor:frfloat;
                               offset_x,offset_y:integer;
                               Pen_,Brush_:tcolor;
                               kanten_ja_nein:boolean);
 procedure zeichne_box(canvas:tcanvas;
                       scalefaktor:frfloat;
                       offset_x,offset_y:integer);
// extrudiere
// save als stl
end;


function Cross(p1, p2: Punkt3D): Punkt3D;
function mach_einheitsvektor(vek:punkt3d):punkt3d;

implementation

// Berechnet den Normalvektor einer Ebene
function Berechne_Normalvektor(p0,p1,p2:punkt3d):punkt3d;
var
v1,v2, erg: Punkt3d; // Zwischenvektoren

  // Bringt den Normalvektor auf EinheitslÃ¤ge 1
  function NormalCut(p:Punkt3d):Punkt3d;
  var
  length: extended;
  begin
   // Wert (LÃ¤nge) des Normalvektors
   length:=Sqrt(p.x*p.x + p.y*p.y + p.z*p.z);
   if length=0 then length:=1;
   p.x:=p.x/length;
   p.y:=p.y/length;
   p.z:=p.z/length;
   result:=p;
  end;

begin
 // Vektor1
 v1.x:=P0.x-P1.x;
 v1.y:=P0.y-P1.y;
 v1.z:=P0.z-P1.z;
 // Vektor2
 v2.x:=P1.x-P2.x;
 v2.y:=P1.y-P2.y;
 v2.z:=P1.z-P2.z;
 // Normalvektor
 erg.x:=v1.y*v2.z - v1.z*v2.y;
 erg.y:=v1.z*v2.x - v1.x*v2.z;
 erg.z:=v1.x*v2.y - v1.y*v2.x;
 // Normalvektor mit LÃ¤nge 1 (Einheitsvektor)
 result:=NormalCut(erg);
end;


// Macht aus dem Vektor vek einen Einheitsvektor , normiert auf Länge 1
function mach_einheitsvektor(vek:punkt3d):punkt3d;
var betrag:extended;
begin
 betrag:=sqrt(sqr(vek.x)+sqr(vek.y)+sqr(vek.z));
 result.x:=vek.x/betrag;
 result.y:=vek.y/betrag;
 result.z:=vek.z/betrag;
end;

// Kreuzprodukt
function Cross(p1, p2: Punkt3D): Punkt3D; // P1,P2 sind Vektoren
begin
 Result.x:= p1.y*p2.z - p1.z*p2.y;
 Result.y:= p1.z*p2.x - p1.x*p2.z;
 Result.z:= p1.x*p2.y - p1.y*p2.x;
end;

///// Dreieck /////
procedure Tdreieck.berechne_normalvektor;
var v1,v2:punkt3d;
begin
 v1.x:=(b.x-a.x);
 v1.y:=(b.y-a.y);
 v1.z:=(b.z-a.z);
 v2.x:=(c.x-b.x);
 v2.y:=(c.y-b.y);
 v2.z:=(c.z-b.z);
 nv:=cross(v1,v2);
 nv:=mach_einheitsvektor(nv);
end;

///// Polygone /////

// die Drehrichtung berechnen
// das ganze Polygon durchgehen, nur bei den Bezierpolygonen anwenden, da es
// bei konkave Polygonen mit zu vielen aufgelösten Punkten sonst schnell zu
// falschen Richtungsangaben kommt
// eventuell noch ne bessere Lösung suchen !!!!!
procedure TGlyph_polygon.berechne_drehrichtung;
var p0,p1,p2,res:punkt3d;
    p:TPolygon_punkt;
    i:integer;
    z:frfloat;
begin
 z:=0;
 for i:=0 to self.punkte_liste.Count-3 do
  begin
   p0.x:=TPolygon_punkt(self.punkte_liste[i]).x;
   p0.y:=TPolygon_punkt(self.punkte_liste[i]).y;
   p0.z:=0;
   p1.x:=TPolygon_punkt(self.punkte_liste[i+1]).x;
   p1.y:=TPolygon_punkt(self.punkte_liste[i+1]).y;
   p1.z:=0;
   p2.x:=TPolygon_punkt(self.punkte_liste[i+2]).x;
   p2.y:=TPolygon_punkt(self.punkte_liste[i+2]).y;
   p2.z:=0;
   res:=Berechne_Normalvektor(p0,p1,p2);
   z:=z+res.z;
  end;

 if z<0 then self.Drehrichtung:=rechts
        else self.Drehrichtung:=links;
end;

// Ein Bezierpolygon in ein Konturpolygon wandeln
// Bezier-Stützpunkte auflösen und durch berechnente Konturpunkte ersetzen
// Das Ganze in Result zurückgeben
function TGlyph_polygon.glaette(aufloesung:integer):TGlyph_polygon;
var np,npi,i,j:integer;
    sl:tstringlist;
    xb,yb, // Bezier-Input
    c,     // Koeffizieneten
    xint,yint:doubarray; // Feine Kontur
    punkt,punkt1,punkt_neu:TPolygon_punkt;
    p1,p2,p_neu,b1,b2,b3:TPolygon_punkt;
    abschnittliste:tlist;
    modliste:tlist;
begin
 result:=TGlyph_polygon.create;

 // Modefiziert Punktliste erzeugen, wo zwischen 2 Bezierpunkten ein Konturpunkt eingefügt wird
 // Mann hat dann die Bezierpunkte vereinzelt , mit je einem Konturpunkt davor und danach
 modliste:=tlist.Create;
 for i:=0 to self.punkte_liste.Count-2 do
  begin
   p1:=TPolygon_punkt(self.punkte_liste[i]);
   p2:=TPolygon_punkt(self.punkte_liste[i+1]);
   if (p1.KB_eigenschaft=konturpunkt) or (p1.KB_eigenschaft=bezierpunkt)
    then
     begin
      p_neu:=TPolygon_punkt.Create;
      p_neu.x:=p1.x;
      p_neu.y:=p1.y;
      p_neu.KB_eigenschaft:=p1.KB_eigenschaft;
      p_neu.SEZ_eigenschaft:=p1.SEZ_eigenschaft;
      modliste.Add(p_neu);
     end;
   // Zwischenpunkt einfügen , wenn zwei Bezierpunkt in Folge kommen
   if (p1.KB_eigenschaft=bezierpunkt) and (p2.KB_eigenschaft=bezierpunkt)
    then
     begin
      p_neu:=TPolygon_punkt.Create;
      p_neu.x:=(p1.x+p2.x)/2;
      p_neu.y:=(p1.y+p2.y)/2;
      p_neu.KB_eigenschaft:=konturpunkt;
      p_neu.SEZ_eigenschaft:=zwischenpunkt; // Default
      modliste.Add(p_neu);
     end;
  end;
 // letzten Punkt noch anfügen
 p_neu:=TPolygon_punkt.Create;
 p_neu.x:=p2.x;
 p_neu.y:=p2.y;
 p_neu.KB_eigenschaft:=p2.KB_eigenschaft;
 p_neu.SEZ_eigenschaft:=p2.SEZ_eigenschaft;
 modliste.Add(p_neu);

 // Nun die Bezierauswertung (glätten)
 for i:=0 to modliste.Count-1 do
  begin
   p1:=TPolygon_punkt(modliste[i]);
   if i<modliste.Count-1
    then p2:=TPolygon_punkt(modliste[i+1])
    else p2:=TPolygon_punkt(modliste[0]);
   if (p1.KB_eigenschaft=konturpunkt)
    then
     begin
      p1.SEZ_eigenschaft:=zwischenpunkt;
      result.punkte_liste.Add(p1);
     end;

   if (p1.KB_eigenschaft=bezierpunkt)
    then
     begin
      // Bezierpunkte auflösen und neue Punkte anfügen
      b1:=TPolygon_punkt(modliste[i-1]);
      b2:=TPolygon_punkt(modliste[i]);
      if i<modliste.Count-1
       then
        b3:=TPolygon_punkt(modliste[i+1])
       else
        b3:=TPolygon_punkt(modliste[0]);
      np:=3; // 3 Punkte für quadratisches Bezier
      setlength(xb,np);
      setlength(yb,np);
      xb[0]:=b1.x;  yb[0]:=b1.y;
      xb[1]:=b2.x;  yb[1]:=b2.y;
      xb[2]:=b3.x;  yb[2]:=b3.y;
      // berechnen der Koeffizienten
      Bezier_Koeff(np,c);
      // Anzahl der gwünschten Konturensegmente (Auflösung)
      npi:=aufloesung;
      // berechnen der feinen Kurvenpunkte
      kurven_punkte(np,xb,yb,c,npi,xint,yint);
      for j:=1 to length(xint)-2 do
       begin
        punkt_neu:=TPolygon_punkt.Create;
        punkt_neu.x:=xint[j];
        punkt_neu.y:=yint[j];
        punkt_neu.KB_eigenschaft:=konturpunkt;
        punkt_neu.SEZ_eigenschaft:=zwischenpunkt; // Default
        result.punkte_liste.Add(punkt_neu);
       end;
     end;
  end;

 // Kann nun weg
 modliste.Free;
end;

// Rotiert die Punktliste
procedure TGlyph_polygon.rotiere_punkteliste_rechts;
var p:pointer;
    punkt:TPolygon_punkt;
    i:integer;
begin
 // den letzten Punkt merken
 p:=self.punkte_liste[self.punkte_liste.Count-1];
 // letzten Eintrag löschen
 self.punkte_liste.Delete(self.punkte_liste.Count-1);
 // Pointer des Punktes vorne einfügen
 self.punkte_liste.Insert(0,p);

 // Eigenschaften von Startpunk , Endpunkt , Zwischenpunkt anpassen
 for i:=0 to self.punkte_liste.Count-1 do
  begin
   punkt:=TPolygon_punkt(self.punkte_liste[i]);
   punkt.SEZ_eigenschaft:=zwischenpunkt; // erst mal Default
   if i=0 then punkt.SEZ_eigenschaft:=startpunkt;
   if i=self.punkte_liste.Count-1 then punkt.SEZ_eigenschaft:=endpunkt;
  end;
end;

// Zeichnent das Polygon auf den Canvas
procedure TGlyph_polygon.zeichne(canvas:tcanvas;
                                 scale:frfloat;
                                 offset_x,offset_y:integer;
                                 punktgroesse:integer;
                                 punkte_zeichnen:boolean);
var i:integer;
    x1,y1,x2,y2:frfloat;
    oldbrushcolor,aktuell_color:tcolor;
    oldpencolor:tcolor;
    tex:string;
    kb:TKB_eigenschaft; //Konturpunkt,Bezierpunkt
    sez:TSEZ_eigenschaft; //Startpunkt,Endpunkt,Zwischenpunkt

    // wenn ein Punkt exakt auf der Kontur liegt
    // Rechteck zeichnen
     procedure POnCurve(x,y:frfloat;color:tcolor);
     var
      Bounds:TRect;
     begin
      oldbrushcolor:=canvas.Brush.Color;
      canvas.Brush.Color:=color;
      bounds.BottomRight.x:=round(x+punktgroesse);
      bounds.BottomRight.y:=round(y+punktgroesse);
      bounds.TopLeft.x:=bounds.BottomRight.x-2*punktgroesse;
      bounds.TopLeft.y:=bounds.BottomRight.y-2*punktgroesse;
      canvas.Rectangle(Bounds);
      canvas.Brush.Color:=oldbrushcolor;
     end;

     // Wenn es eine Bezier-Stützstelle ist
     // Kreis zeichnen
     procedure PNotOnCurve(x,y:frfloat;color:tcolor);
     var
      Bounds:TRect;
     begin
      oldbrushcolor:=canvas.Brush.Color;
      canvas.Brush.Color:=color;
      bounds.BottomRight.x:=round(x+punktgroesse);
      bounds.BottomRight.y:=round(y+punktgroesse);
      bounds.TopLeft.x:=bounds.BottomRight.x-2*punktgroesse;
      bounds.TopLeft.y:=bounds.BottomRight.y-2*punktgroesse;
      canvas.Ellipse(Bounds);
      canvas.Brush.Color:=oldbrushcolor;
     end;

begin
 canvas.moveto(0,0);
 // Startpunkt festhalten
 x1:=TPolygon_punkt(self.punkte_liste[0]).x*scale;
 y1:=TPolygon_punkt(self.punkte_liste[0]).y*scale;

 // Punkte malen
 if punkte_zeichnen
  then
   for i:=0 to self.punkte_liste.Count-1 do
    begin
     x2:=TPolygon_punkt(self.punkte_liste[i]).x*scale;
     y2:=TPolygon_punkt(self.punkte_liste[i]).y*scale;
     // Farbe festlegen
     sez:=TPolygon_punkt(self.punkte_liste[i]).SEZ_eigenschaft;
     kb:=TPolygon_punkt(self.punkte_liste[i]).KB_eigenschaft;
     // Punkt ein Startpunkt ist
     if sez=startpunkt then aktuell_color:=clyellow;
     if sez=endpunkt then aktuell_color:=cllime;
     if sez=zwischenpunkt then aktuell_color:=clgreen;
     if (sez=zwischenpunkt) and (kb=bezierpunkt) then aktuell_color:=clred;

     // Wenn es ein Bezierpunkt ist , dann rund , sonst Quadrat
     if kb=bezierpunkt
      then
       PNotOnCurve(x2+offset_x,canvas.Height-y2-offset_y,aktuell_color)
      else
       POnCurve(x2+offset_x,canvas.Height-y2-offset_y,aktuell_color);
    end;

 // Linien zeichnen
 oldpencolor:=canvas.Pen.Color;
 // Je nach Drehrichtung , Farbe einstellen
 if self.Drehrichtung=rechts then canvas.pen.Color:=clred
                             else canvas.pen.Color:=clblue;
 canvas.MoveTo(round(x1)+offset_x,canvas.Height-round(y1)-offset_y);
 for i:=1 to self.punkte_liste.Count-1 do
  begin
   x2:=TPolygon_punkt(self.punkte_liste[i]).x*scale;
   y2:=TPolygon_punkt(self.punkte_liste[i]).y*scale;
   canvas.lineto(round(x2)+offset_x,canvas.Height-round(y2)-offset_y);
  end;
 // letzte Linie zurück zum Startpunkt
 canvas.lineto(round(x1)+offset_x,canvas.Height-round(y1)-offset_y);
 canvas.Pen.Color:=oldpencolor;

end;

// Polygone-Objekt erzeugen
constructor TGlyph_polygon.create;
begin
 // Punkteliste anlegen
 punkte_liste:=tlist.Create; // TFont_punkt;
end;

// Polygone-Objekt löschen
destructor TGlyph_polygon.destroy;
var i:integer;
begin
 // Die Punktliste löschen
 for i:=0 to punkte_liste.count-1 do
  if TPolygon_punkt(punkte_liste[i]) <> nil
   then TPolygon_punkt(punkte_liste[i]).free;
 punkte_liste.free;
 inherited destroy;
end;


///// 3D Glyph /////

// die Bounding-Box um das aufgelöste Glyph zeichnen
procedure T3D_Glyph.zeichne_box(canvas:tcanvas;
                                scalefaktor:frfloat;
                                offset_x,offset_y:integer);
var old_pen:tcolor;
begin
 old_pen:=canvas.Pen.Color;
 canvas.Pen.Color:=cllime;;
 canvas.MoveTo(round(box.min.x*scalefaktor)+offset_x,
               canvas.Height-round(box.min.y*scalefaktor)-offset_y);
 canvas.lineto(round(box.min.x*scalefaktor)+offset_x,
               canvas.Height-round(box.max.y*scalefaktor)-offset_y);
 canvas.lineTo(round(box.max.x*scalefaktor)+offset_x,
               canvas.Height-round(box.max.y*scalefaktor)-offset_y);
 canvas.lineto(round(box.max.x*scalefaktor)+offset_x,
               canvas.Height-round(box.min.y*scalefaktor)-offset_y);
 canvas.lineTo(round(box.min.x*scalefaktor)+offset_x,
               canvas.Height-round(box.min.y*scalefaktor)-offset_y);

 canvas.Pen.Color:=old_pen;
end;

// Die Polygone mit aufgelösten Bezierpunkten zeichnen
procedure T3D_Glyph.zeichne_fein_polys(canvas:tcanvas;
                                       scalefaktor:frfloat;
                                       offset_x,offset_y:integer;
                                       punkte_ja_nein:boolean);
var i:integer;
    feinPoly:TGlyph_polygon;
    old_pen,old_brush:tcolor;
begin
 old_pen:=canvas.Pen.Color;
 old_brush:=canvas.Brush.Color;
 // Die groben Glyph-Polygone zeichnen
 for i:=0 to self.fein_Polys.Count-1 do
  begin
   feinPoly:=TGlyph_polygon(self.fein_Polys[i]);
   feinPoly.berechne_drehrichtung;
   feinPoly.zeichne(Canvas,scalefaktor,offset_x,offset_y,4,punkte_ja_nein);
  end;
 canvas.Pen.Color:=old_pen;
 canvas.Brush.Color:=old_brush;
end;

// Die Grob-Polygone (Fontdaten) zeichnen
procedure T3D_Glyph.zeichne_grob_polys(canvas:tcanvas;
                                       scalefaktor:frfloat;
                                       offset_x,offset_y:integer;
                                       punkte_ja_nein:boolean);
var i:integer;
    GrobPoly:TGlyph_polygon;
    old_pen,old_brush:tcolor;
begin
 old_pen:=canvas.Pen.Color;
 old_brush:=canvas.Brush.Color;
 // Die groben Glyph-Polygone zeichnen
 for i:=0 to self.Grob_Polys.Count-1 do
  begin
   GrobPoly:=TGlyph_polygon(self.Grob_Polys[i]);
   GrobPoly.berechne_drehrichtung;
   GrobPoly.zeichne(Canvas,scalefaktor,offset_x,offset_y,4,punkte_ja_nein);
  end;
 canvas.Pen.Color:=old_pen;
 canvas.Brush.Color:=old_brush;
end;

// Die 2D-Dreicke auf ein Canvas zeichnen
procedure T3D_Glyph.zeichne_dreiecke_2D(canvas:tcanvas;
                                        scalefaktor:frfloat;
                                        offset_x,offset_y:integer;
                                        Pen_,Brush_:tcolor;
                                        kanten_ja_nein:boolean);
var i:integer;
    dreieck:tdreieck;
    points:array of TPoint;
    old_pen,old_brush:tcolor;
begin
 old_pen:=canvas.Pen.Color;
 old_brush:=canvas.Brush.Color;
// canvas.MoveTo(0,0);
 if kanten_ja_nein
  then canvas.Pen.Color:=pen_
  else canvas.Pen.Color:=brush_;
 canvas.Brush.Color:=Brush_;
 setlength(points,3);
 for i:=0 to self.Dreiecklist2D.Count-1 do // Anzahl der Dreiecke
  begin
   dreieck:=Tdreieck(self.Dreiecklist2D[i]);
   points[0].x:=round(dreieck.A.x*scalefaktor)+offset_x;
   points[0].y:=Canvas.Height-round(dreieck.A.y*scalefaktor)-offset_y;
   points[1].x:=round(dreieck.B.x*scalefaktor)+offset_x;
   points[1].y:=Canvas.Height-round(dreieck.B.y*scalefaktor)-offset_y;
   points[2].x:=round(dreieck.C.x*scalefaktor)+offset_x;
   points[2].y:=Canvas.Height-round(dreieck.C.y*scalefaktor)-offset_y;
   canvas.Polygon(points);
  end;
 canvas.Pen.Color:=old_pen;
 canvas.Brush.Color:=old_brush;
end;

// Erzeugt ein BeRo kompatibles Format
function T3D_Glyph.mach_bero_polys:TBeRoTriangulationPolygons;
var i,j:integer;
    polygon:TGlyph_polygon;
    punkt:TPolygon_punkt;
begin
 setlength(result,self.Fein_Polys.Count); // Anzahl der Polygone setzten
 for i:=0 to self.Fein_Polys.Count-1 do
  begin
   polygon:=TGlyph_polygon(self.Fein_Polys[i]);
   setlength(result[i],polygon.punkte_liste.Count); // Anzahl der Punkte setzten
   // alle Polygonpunkte durchjukeln
   for j:=0 to polygon.punkte_liste.Count-1 do
    begin
     punkt:=TPolygon_punkt(polygon.punkte_liste[j]);
     // Koordinaten übertragen
     result[i,j].x:=punkt.x;
     result[i,j].y:=punkt.y;
    end; // for j
  end; // for i
end;

// Brechnet die Bounding-Box um das Glyph , Z wird erst mal auf 0 gehalten
Procedure T3D_Glyph.berechne_box;
var polygon:TGlyph_polygon;
    punkt:TPolygon_punkt;
    i,j,x,y:integer;
begin
 self.box.min.x:=10000000;
 self.box.min.y:=10000000;
 self.box.min.z:=0;
 self.box.max.x:=-10000000;
 self.box.max.y:=-10000000;
 self.box.max.z:=0;
 // alle Glymphpunkte durchjukeln
 for j:=0 to _glyph^.outline.n_points-1 do
  begin
   x:=_glyph^.outline.points^[j].x;
   y:=_glyph^.outline.points^[j].y;
   // Min
   if x<self.box.min.x then self.box.min.x:=x;
   if y<self.box.min.y then self.box.min.y:=y;
   // Max
   if x>self.box.max.x then self.box.max.x:=x;
   if y>self.box.max.y then self.box.max.y:=y;
  end;
end;

// 3D Glyph-Objekt erzeugen
constructor T3D_Glyph.create(const glyph:TFreeTypeGlyph);
var
    i,j,pc,start_,end_:integer;
    polygon,polygone_:TGlyph_polygon;
    polypunkt:TPolygon_punkt;
begin
 // Sammelliste für die Polygone erzeugen , Gobe Liste mit Bezierpunkten
 Grob_Polys:=tlist.Create;
 // Fein_polygonliste alle Bezier durch berechnete Konturpunkt ersetzt
 Fein_Polys:=tlist.Create;
 // Pointer auf die Glyph-Daten
 _glyph:=glyph.Data.z;

 // 2D Dreieckliste anlegen
 Dreiecklist2D:=tlist.Create; //TDreieck platt in 2D , Z-Koordinate = 0 , nur den vorderen Deckel

 // 3D Dreieckliste anlegen
 Dreiecklist3D:=tlist.Create; //TDreieck Körper extrudiert

 // Box berechnen
 berechne_box;

 // Anzahl der Polygone , nen A hat 2 , ne 8 hat 3 Polygone
 anzahl_polygone:=_glyph^.outline.n_contours;
 // Das Glyph auswerten und davon Polygone bauen
 start_:=0;
 // Die Fontpolygone separieren und ablegen
 for pc:=0 to anzahl_polygone-1 do
  begin
   // Liste mit den Polygon-Enden
   end_:=_glyph^.outline.conEnds^[pc];
   // Erzeuge ein Polygon
   polygone_:=TGlyph_polygon.create;
   // an die Font_polygonliste des Glyph-Objektes anfügen
   Grob_Polys.Add(polygone_);
   // Die Punkte rausziehen und an das poly.punkte_liste anfügen
   for j:=start_ to end_ do
    begin
     // Punktobjekt erzeugen
     polypunkt:=TPolygon_punkt.Create;
     // koordinaten holen
     polypunkt.x:=_glyph^.outline.points^[j].x;
     polypunkt.y:=_glyph^.outline.points^[j].y;

     // KB_Eigenschaft des Punktes , auf Kontur oder Bezier
     if (_glyph^.outline.flags^[j] and TT_Flag_On_Curve) <>0
      then
       polypunkt.kb_eigenschaft:=Konturpunkt
      else
       polypunkt.kb_eigenschaft:=Bezierpunkt;

     // SEZ_Eigenschaft , Startpunkt , Endpunkt , Zwischenpunkt
     polypunkt.SEZ_eigenschaft:=zwischenpunkt; // Default
     if j=start_ then polypunkt.SEZ_eigenschaft:=startpunkt;
     if j=end_ then polypunkt.SEZ_eigenschaft:=endpunkt;

     // Punkt an die Liste anfügen
     polygone_.punkte_liste.Add(polypunkt);
    end;
   start_:=end_+1;
  end;
end;

// 3D Glyph-Objekt freigeben
destructor T3D_Glyph.destroy;
var i:integer;
begin
 // Die Font-Polygonliste löschen
 for i:=0 to Grob_Polys.count-1 do
  if TGlyph_polygon(Grob_Polys[i]) <> nil
   then TGlyph_polygon(Grob_Polys[i]).free;
 Grob_Polys.free;

 // Kontur_polygonliste löschen
 for i:=0 to Fein_Polys.count-1 do
  if TGlyph_polygon(Fein_Polys[i]) <> nil
   then TGlyph_polygon(Fein_Polys[i]).free;
 Fein_Polys.free;

 // 2D Dreieckliste löschen
 for i:=0 to Dreiecklist2D.count-1 do
  if TDreieck(Dreiecklist2D[i]) <> nil
   then TDreieck(Dreiecklist2D[i]).free;
 Dreiecklist2D.free;

 // 3D Dreieckliste löschen
 for i:=0 to Dreiecklist3D.count-1 do
  if TDreieck(Dreiecklist3D[i]) <> nil
   then TDreieck(Dreiecklist3D[i]).free;
 Dreiecklist3D.free;

 inherited destroy;
end;

end.

